package com.mhsoftware.cdaily.support.security;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.HttpURLConnection;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Proxy;
import java.net.SocketAddress;
import java.net.URL;
import java.util.List;
import java.util.Locale;
import java.util.Properties;
import java.util.ResourceBundle;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.servlet.ServletContext;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;

import com.MHSoftware.net.TrustModifier;
import com.MHSoftware.string.StringUtil;
import com.mhsoftware.cdaily.business.User;
import com.mhsoftware.cdaily.support.WebApp;


/**
 * This method provides a cookie based SingleSignon authentication provider. 
 * It takes a specified set of cookies from the request, and sends them to 
 * a specified URL. The code specified by the URL will return optional elements 
 * that indicate whether the authentication was successful. 
 *  
 * The following elements are defined in /WEB-INF/singlesignon.properties 
 *  
 * <dl> 
 * <dt>cookieRegex - required
 * <dd>A regular expression of cookies that exist on the request that should be sent to the AuthURL. This will be done using cookieName.matches(regex).
 *  
 * <dt>AuthURL - Required 
 * <dd>The URL the required cookies should be sent to. 
 *  
 * <dt>AuthRequestType - optional 
 * <dd>Http Request Type, e.g. GET (default), POST, HEAD 
 *  
 * <dt>errorRegex - optional 
 * <dd>Regular expression to match an error message. If this expression is matched, an error will be assumed and the result 
 * will be returned as an error message. 
 *  
 * <dt>userRegex - required 
 * <dd>Regular expression that when applied to the response body, will return the name of the authenticated user. 
 *  
 * <dt>fullnameRegex - optional 
 * <dd>Regular expression that when applied to the response body, will return the full name of the authenticated user. 
 *  
 * <dt>emailRegex - optional 
 * <dd>Regular expression that when applied to the response body, will return the email address of the authenticated user. 
 *  
 * <dt>groupsRegex - optional  
 * <dd>Regular expression that returns a comma delimited list of groups the user is a member of. 
 *  
 * <dt>Debug - Optional, Default False
 * <dd>Yes: Enable debugging output to stdout. 
 *  
 * <dt>ProxyServer, ProxyPort - Optional
 * <dd>Proxy Server name (or address) and port to use for the HttpConnection. 
 *  
 * <dt>IgnoreSSLErrors - Default no 
 * <dd>Yes: Ignore any SSL Certificate errors.
 *  
 * </dl>
 *  
 * @since 3.4.15 
 * 
 * @author gsexton (6/12/2012)
 */
public class AuthProviderCookieSignon extends AuthProviderSingleSignon {

    public static void init(ServletContext sc) {
        context = sc;
        provider_flags = AuthenticationProvider.DOES_AUTO_ADD | AuthenticationProvider.SINGLE_SIGNON;
        getConfiguration();	// This will turn on debug mode when the property file is read.
        if (isDebugMode()) {
            System.out.println("Using CookieSignon AuthenticationProvider!");
        }
    }


    /** 
     * If debugging is enabled, print the fatal list to stdout. 
    */
    private void printFatalList() {
        if (!isDebugMode()) {
            return;
        }
        if (lr != null) {
            List<String>lst = lr.getFatalList();
            for (String s: lst) {
                System.out.println("FATAL: " + s);
            }
        }
    }

    /**
     * This method sends the authentication request to the authentication source.
     * 
     * @param req 
     * @param rb 
     * 
     * @return String 
     */
    private String sendAuthenticationRequest(HttpServletRequest req, ResourceBundle rb) {
        /**
         * OK, we need to do our http request.
         */
        Cookie[] cookies = req.getCookies();
        if (cookies == null) {
            lr.getFatalList().add(rb.getString("Security.AuthenticationFailure") + " No Cookies on request.");
            if (isDebugMode()) {
                printFatalList();
            }
            return null;
        }
        Properties pConfig = getConfiguration();
        String sCookieRegex = pConfig.getProperty("cookieRegex");

        StringBuilder sb = new StringBuilder();
        for (Cookie cookie: cookies) {
            if (isDebugMode()) {
                System.out.println("Examining cookie: " + cookie.getName() + " it " + (cookie.getName().matches(sCookieRegex) ? " matches" : "doesn't match."));
            }
            if (cookie.getName().matches(sCookieRegex)) {
                if (sb.length() > 0) {
                    sb.append(';');
                }
                try {
                    sb.append(cookie.getName()).append('=').append(java.net.URLEncoder.encode(cookie.getValue(), "ISO-8859-1"));
                } catch (java.io.UnsupportedEncodingException  ignore) {}
            }
        }
        if (sb.length() == 0) {
            lr.getFatalList().add(rb.getString("Security.AuthenticationFailure") + " No Cookies match expression: " + sCookieRegex);
            printFatalList();
            return null;
        }
        if (isDebugMode()) {
            System.out.println("AuthURL=" + pConfig.getProperty("AuthURL"));
            System.out.println("Cookie String=" + sb.toString());
        }
        HttpURLConnection conn = null;
        try {
            URL url = new URL(pConfig.getProperty("AuthURL"));
            if (pConfig.getProperty("ProxyServer") != null) {
                SocketAddress sa = new InetSocketAddress(InetAddress.getByName(pConfig.getProperty("ProxyServer")), Integer.parseInt(pConfig.getProperty("ProxyPort", "80")));
                Proxy p = new Proxy(Proxy.Type.HTTP, sa);
                if (isDebugMode()) {
                    System.out.println("Using Proxy Server: "+pConfig.getProperty("ProxyServer")+" Port:"+pConfig.getProperty("ProxyPort"));
                }
                conn = (HttpURLConnection)url.openConnection(p);
            } else {
                conn = (HttpURLConnection)url.openConnection();
            }
            conn.setRequestMethod(pConfig.getProperty("AuthRequestType", "GET"));
            conn.setConnectTimeout(5000);
            conn.setReadTimeout(5000);
            conn.setDoInput(true);
            conn.addRequestProperty("Cookie", sb.toString());
            conn.addRequestProperty("Cache-Control", "no-cache");
            conn.addRequestProperty("Pragma", "no-cache");
            conn.addRequestProperty("User-Agent", WebApp.PRODUCT_NAME + " v" + WebApp.PRODUCT_VERSION + " Cookie Signon Authenticator");
            if ("yes".equalsIgnoreCase(pConfig.getProperty("IgnoreSSLErrors","No"))) {
                if (isDebugMode()) {
                    System.out.println("Setting option to ignore SSL Handshake Errors.");
                }
                try { 
                    TrustModifier.relaxHostChecking(conn);
                } catch (Exception ex) {
                    throw new IOException(ex);
                }
            }
            conn.connect();

            BufferedReader br  = new BufferedReader(new InputStreamReader(conn.getInputStream(), java.nio.charset.Charset.forName("UTF-8")));
            sb = new StringBuilder();
            String line;
            while ((line = br.readLine()) != null) {
                sb.append(line).append('\n');
            }
            br.close();
            conn.disconnect();
            if (isDebugMode()) {
                System.out.println("Connection Result=" + sb.toString());
            }
            return sb.toString();
        } catch (IOException ioe) {
            lr.getFatalList().add(ioe.toString());
            System.out.println(ioe);
        }
        printFatalList();
        return null;
    }

    private String extractValue(final String regex, final String findIn) {
        Pattern p = Pattern.compile(regex);
        Matcher m = p.matcher(findIn);
        String sResult = null;
        if (m.find()) {
            sResult = m.group(1);
        }
        return sResult;
    }


    private void processResponse(final String sResponse, ResourceBundle rb, Locale lc) {
        if (isDebugMode()) {
            System.out.println("AuthProviderCookieSignon - processResponse(sResponse=[" + sResponse + "])");
        }
        Properties pConfig = getConfiguration(),pResult = new Properties();
        final String[] aExpressions = { "errorRegex", "userRegex", "fullnameRegex", "groupsRegex", "emailRegex" };
        for (String s: aExpressions) {
            String sExpr = pConfig.getProperty(s);
            if (sExpr != null) {
                String sValue = extractValue(sExpr, sResponse);
                if (sValue != null) {
                    pResult.put(s, sValue);
                }
            }
        }
        if (isDebugMode()) {
            System.out.println("Response contains: " + pResult.toString());
        }
        if (pResult.getProperty(aExpressions[0]) != null) {
            // There is an error.
            lr.getFatalList().add(pResult.getProperty(aExpressions[0]));
            printFatalList();
            return;
        }
        String sUser = pResult.getProperty(aExpressions[1]);
        if (sUser == null) {
            lr.getFatalList().add(rb.getString("Security.AuthenticationFailure"));
        } else {
            /** The template create user forces the user name to lower case, so we need to do that here as well. */
            sUser = sUser.toLowerCase(lc);
            User u = User.getUserByName(sUser);
            if (u == null) {
                u = createNewUser(sUser, pResult.getProperty("fullnameRegex"), pResult.getProperty("emailRegex"), pResult.getProperty("groupsRegex"));
            } else {
                if (u.isEnabled()) {
                    updateUser(u, pResult.getProperty("fullnameRegex"), pResult.getProperty("emailRegex"), pResult.getProperty("groupsRegex"));
                }
            }
            lr.setUser(u);
        }
    }

    public LoginResult authenticateUser(HttpServletRequest req, ResourceBundle rb, Locale lc) {
        if (isDebugMode()) {
            System.out.println("CookieSignon.authenticateUser(HttpServletRequest,ResourceBundle,Locale) called.");
        }
        lr = new LoginResult();
        String sResponse = sendAuthenticationRequest(req, rb);
        if (StringUtil.isEmpty(sResponse)) {
            return lr;
        }
        processResponse(sResponse, rb, lc);
        User u = lr.getUser();
        if (u == null) {
            /* 
                Login Failed. Not sure what to do here. 
            */
        } else {
            if (verifyLoginPreconditions(u, rb).isEmpty()) {
                processSuccessfulLogin(u, rb, lc);
            } else {
                if (isDebugMode()) {
                    System.out.println("AuthProviderCookieSignon.authenticateUser(req,rb,lc) - verifyLoginPreconditions() returned false.");
                }
                processFailedLogin(u, rb);
            }
        }
        printFatalList();
        return lr;
    }

}
