package com.mhsoftware.cdaily.support.security;
/*
	Connect Daily Web Calendar Software

	Copyright 2002-2006, MH Software, Inc. All Rights Reserved
	5023 W 120th Ave #311, Broomfield CO 80020
	+1 303 438 9585

	Last Edit By: "$Author: gsexton $"
	Last Checkin: "$Date: 2014/09/16 22:16:36 $"
	Revision #  : "$Revision: 1.35 $"

*/


import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.sql.SQLException;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.servlet.ServletContext;

import com.MHSoftware.net.TrustingLDAPSSLFactory;
import com.MHSoftware.regex.RegexConstants;
import com.MHSoftware.string.StringUtil;
import com.mhsoftware.cdaily.business.User;
import com.mhsoftware.cdaily.business.groups;
import com.mhsoftware.cdaily.business.groupusers;
import com.mhsoftware.cdaily.support.WebApp;
import com.novell.ldap.LDAPAttribute;
import com.novell.ldap.LDAPAttributeSet;
import com.novell.ldap.LDAPConnection;
import com.novell.ldap.LDAPEntry;
import com.novell.ldap.LDAPException;
import com.novell.ldap.LDAPJSSESecureSocketFactory;
import com.novell.ldap.LDAPReferralException;
import com.novell.ldap.LDAPSearchResults;
import com.novell.ldap.LDAPSocketFactory;

/**
 * Implements AuthenticationProvider via LDAP authentication.
 * SSL support for LDAP is required.
 * <BR><BR>
 * Using this class, you can authenticate to any LDAP directory
 * source. This includes NetWare, Microsoft ActiveDirectory,
 * and UNIX LDAP servers.
 * <BR><BR>
 * <H1>Known or suspected issues</H1>
 * I have seen some strange behavior on Linux. When the
 * authentication fails, subsequent SSL requests will block.
 * I'm thinking that it's a Linux JDK issue. The same code
 * run from a Windows workstation against the same LDAP
 * server works correctly. 
 * @author George Sexton
 */
public class AuthProviderLDAP extends AuthenticationProvider {

	protected static final int ldapVersion  = LDAPConnection.LDAP_V3;
	protected static final String LDAP_CHARSET="UTF-8";
	private static final String SAMACCOUNTNAME="SAMAccountName";
	private static final String UPN="userPrincipalName";
	private static final String DEFAULT_SERVER_PREFNAME="DefaultLDAPServerName";
	private static final String CHECK_SSL_CERT_PROP="CheckSSLCertificate";


	/** Property bag to hold the LDAP data connection
	 * information.
	 *
	 */
	protected static Properties propBag=null;
	private static Credentials lookupCredentials=null;

	public static void init(ServletContext sc){
		context=sc;
		provider_flags=AuthenticationProvider.DIRECTORY_AVAILABLE |
			AuthenticationProvider.DOES_AUTO_ADD;
		init(sc.getResourceAsStream("/WEB-INF/ldap.properties"));
	}

	private static class Credentials {
		public String userName;
		public String password;

		public Credentials(String user,String pw) {
			this.userName=user;
			this.password=pw;
		}
	}


	private static class UnboundConnection {
		public LDAPConnection lc;
		public LDAPException bindException=null;
		public String additionalErrorInfo=null;

		public UnboundConnection(LDAPConnection lc) {
			this.lc=lc;
		}

		public boolean isBound() {
			return false;
		}
		public void disconnect(){
			if (lc!=null) {
				try {
					lc.disconnect();
				} catch (LDAPException lc) {	//NOPMD
					//	Ignore this.
				}
				lc=null;
			}
		}
	}

	/** 
	 * Guess the distinguished name using the user name, and some of the prefix/suffix user values from config. 
	 * @param UserID, the undistinguished name. E.G. SamAccountName 
	*/
	static String guessDN(String UserID){
		StringBuilder loginDN=new StringBuilder(64);
		String sPrefix=propBag.getProperty("PrefixUserWith");

		if (sPrefix!=null && !UserID.startsWith(sPrefix)) {
			loginDN.append(sPrefix);
		}

		loginDN.append(UserID.trim());

		String sSuffix=propBag.getProperty("SuffixUserWith");

		if (sSuffix!=null && !UserID.endsWith(sSuffix)) {
			loginDN.append(sSuffix);
		}

		return loginDN.toString();
	}

	private static class BoundConnection extends UnboundConnection {
		public String serverName;
		public Credentials credentials;
		private Map<String,String> dnCache=new HashMap<String,String>();        

		public BoundConnection(UnboundConnection uc,String serverName,Credentials credentials) {
			super(uc.lc);
			this.serverName=serverName;
			this.credentials=credentials;
		}
		/**
		 * Get the distinguished name for the specified user ID. If we're bound, or there are lookup 
		 * credentials specified, the dn will be looked up in a query, otherwise it will be guessed. 
		 * 
		 * @author gsexton (6/30/2012)
		 * 
		 * @param UserID 
		 * 
		 * @return String 
		 */
		public String getDN(String UserID){
			String sResult=dnCache.get(UserID);
			if (sResult==null) {
				if (isBound()) {
					// If we already have a bound connection, do the lookup.
					sResult=lookupDN(UserID);
					if (sResult!=null) {
						dnCache.put(UserID,sResult);
					}
				} else {
					sResult=guessDN(UserID);
				}
			}
			if (isDebugMode()) {
				WebApp.log("getDN(\""+UserID+"\") returns "+sResult);
			}
			return sResult;
		}

		public String toString() {
			StringBuilder sb=new StringBuilder(1024);
			sb.append("BoundConnection {\r\nserverName: ").append(serverName);
			if (credentials==null) {
				sb.append("\r\ncredentials=null");
			} else {
				sb.append("\r\nuserName: ").append(credentials.userName)
				.append("\r\npassword: ").append(credentials.password);
			}
			sb.append("\r\nLDAP Connection: ").append(lc);
			if (lc!=null) {
				sb.append("\r\nLDAP Connection is").append(lc.isConnected()? "" : " not").append(" connected.");
			}
			if (bindException!=null) {
				sb.append("\r\nbindException: ").append(bindException.toString());
			}
			if (additionalErrorInfo!=null) {
				sb.append("\r\nadditionalErrorInfo: "+additionalErrorInfo);
			}
			sb.append("\r\n}\r\n");
			return sb.toString();
		}

		public boolean isBound() {
			if (lc==null) {
				return false;
			} else {
				if (lc.isConnected()) {
					return true;
				} else {
					lc=null;
				}
			}
			return false;
		}

		public String getDSAttribute(final String UserID,final String AttrName){
			if (!isBound()) {
				throw new IllegalStateException("getDSAttribute() when not bound!");
			}
			String sResult=null;
			try {
				String lookupDN=getDN(UserID);
				if (lookupDN!=null) {
					LDAPEntry entry=lc.read(lookupDN,new String[] {AttrName});
					if (entry!=null) {
						LDAPAttribute attr=entry.getAttribute(AttrName);
						if (attr!=null) {
							sResult=attr.getStringValue();
						}
					}
				}
			} catch (LDAPException e) {
				WebApp.log("getDSAttribute()",e);
			}
			if (isDebugMode()) {
				WebApp.log("getDSAttribute(\""+UserID+"\",\""+AttrName+"\") returns "+sResult);
			}
			return sResult;
		}
		private Set<String> getDSAttributeSet(final String UserID,final String AttrName){
			if (!isBound()) {
				throw new IllegalStateException("getDSAttributeSet() called when connection not bound!");
			}
			Set<String> setResult=new HashSet<String>();

			if (isDebugMode()) {
				WebApp.log("\tgetDSAttributeSet(\""+UserID+"\",\""+AttrName+"\");\n");
			}
			String lookupDN=getDN(UserID);
			try {
				if (lookupDN==null) {
					if (isDebugMode()) {
						WebApp.log("getDSAttributeSet(), LookupDN=null!");
					}
				} else {
					if (isDebugMode()) {
						WebApp.log("getDSAttributeSet() LookupDN="+lookupDN);
					}
					LDAPEntry entry=lc.read(lookupDN,new String[] {AttrName});
					if (entry!=null) {
						LDAPAttributeSet attrset=entry.getAttributeSet();
						if (attrset!=null) {
							Iterator itr=attrset.iterator();
							while (itr.hasNext()) {
								LDAPAttribute attr=(LDAPAttribute)itr.next();
								// DOn't ask me how it works. I wrote it using a debugger to figure 
								// out what the hell goes on
								String[] aGroups=attr.getStringValueArray();
								if (aGroups!=null) {
									for (String s : aGroups) {
										setResult.add(s);
									}
								}

							}
						}
					}
				}
			} catch (LDAPException e) {
				WebApp.log("Exception Occurred reading Attribute Set "+AttrName+" for Distinguished Name=["+lookupDN+"]");
				this.bindException=e;
				this.additionalErrorInfo=null;
				WebApp.log("getDSAttributeSet()",e);
			}
			return setResult;
		}



		/**
		 * This  does a lookup of the DN via the userPrincipalName
		 * (or other field) and returns
		 * the DN.
		 * @param UserID The ID of the user to find the DN for.
		 */
		private String lookupDN(String UserID){
			bindException=null;
			additionalErrorInfo=null;
			int iFound=0;
			String sFailureInfo=null;
			String sResult=null;
			String sQueryPart=propBag.getProperty("searchField",UPN);
			String loginDN=guessDN(UserID);
			if (SAMACCOUNTNAME.equalsIgnoreCase(sQueryPart) && loginDN!=null && loginDN.indexOf('\\')>0) {
				loginDN=loginDN.substring(loginDN.indexOf('\\')+1);
			}
			String sDelimiter=propBag.getProperty("ContextDelimiter",";");

			String[] aContexts=propBag.getProperty(serverName+".DefaultContext",
												   propBag.getProperty("DefaultContext","")).split(sDelimiter);
			/*
			if (propBag.getProperty("DefaultContext","")==null) {
				aContexts=new String[1];
			} else {
				
			} 
			*/ 
			if (sQueryPart.length()>0) {
				sQueryPart+="=";
			}
			String searchFilter="("+sQueryPart+loginDN+")";
			for (String sDefaultContext : aContexts) {
				try {
					if (isDebugMode()) {
						WebApp.log("Default Context= ["+sDefaultContext+"] Search Filter=["+searchFilter+"]");
					}
					LDAPSearchResults searchResults =lc.search(  sDefaultContext,
																 LDAPConnection.SCOPE_SUB,
																 searchFilter,
																 new String[] {"distinguishedName"},
																 false);

					if (searchResults==null) {
						if (isDebugMode()) {
							WebApp.log("SearchResults came back null!");
						}
					} else {
						while (searchResults.hasMore()) {
							LDAPEntry nextEntry=null;
							try {
								nextEntry=searchResults.next();
								loginDN=nextEntry.getDN();
								iFound++;
							} catch (LDAPReferralException le) {
								continue;
							}
						}
					}
					switch (iFound) {
					case 1 :
						sResult=loginDN;
						break;
					case 0:
						if (isDebugMode()) {
							WebApp.log("Didn't find the user!");
						}
						break;
					default:
						if (isDebugMode()) {
							WebApp.log("This is serious. We found more than one user!");
						}
						break;
					}
				} catch ( LDAPException e ) {
					bindException=e;
					switch (e.getResultCode()) {
					case LDAPException.NO_SUCH_OBJECT:
						sFailureInfo="Error: No such entry";
						break;
					case LDAPException.INSUFFICIENT_ACCESS_RIGHTS:
						sFailureInfo="Error: Insufficient rights";
						break;
					case LDAPException.ATTRIBUTE_OR_VALUE_EXISTS:
						sFailureInfo="Error: Attribute or value exists";
						break;
					case LDAPException.INVALID_CREDENTIALS:
						sFailureInfo="Error: Invalid Credentials";
						break;
					default:
						sFailureInfo="Error ResultCode=["+e.getResultCode()+"] "+e.resultCodeToString();
						break;
					}
					additionalErrorInfo=sFailureInfo;
					WebApp.log(sFailureInfo,e);
				}
				if (sResult!=null) {
					break;
				}
			}
			return sResult;

		}

	}

	private static int getLDAPPort(){
		if ("I UNDERSTAND THIS IS A MAJOR SECURITY RISK AND ACCEPT FULL RESPONSIBILITY".equals(propBag.getProperty("SSL_DISCLAIMER"))) {
			return LDAPConnection.DEFAULT_PORT;
		} else {
			return LDAPConnection.DEFAULT_SSL_PORT;
		}
	}


	private static UnboundConnection getUnboundConnection(){
		LDAPSocketFactory lsf=null;
		if (getLDAPPort()==LDAPConnection.DEFAULT_SSL_PORT) {
			if ("yes".equalsIgnoreCase(propBag.getProperty(CHECK_SSL_CERT_PROP))) {
				lsf=new LDAPJSSESecureSocketFactory();
			} else {
				lsf=TrustingLDAPSSLFactory.getInstance();
			}
		}
		return new UnboundConnection(lsf==null ? new LDAPConnection() : new LDAPConnection(lsf));
		// return new UnboundConnection(
		// getLDAPPort()==LDAPConnection.DEFAULT_SSL_PORT ? new LDAPConnection(new LDAPJSSESecureSocketFactory()) :  new LDAPConnection()
		// );
	}

	public static void init(String sPropertiesFile){
		if (isDebugMode()) {
			WebApp.log("Using property file "+sPropertiesFile);
		}
		try {
			File f=new File(sPropertiesFile);
			FileInputStream fis=new FileInputStream(f);
			init(fis);
		} catch (IOException ioe) {
			WebApp.log("Error initializing using Property File:"+sPropertiesFile,ioe);
		}
	}

	private static boolean guessUserDN(){
		return "yes".equalsIgnoreCase(propBag.getProperty("GuessDN"));
	}

	public static void init(InputStream is){
		propBag=new Properties();
		final String STARS="\n**************************************************************************\n";
		try {
			propBag.load(is);
			setDebugMode(propBag.getProperty("Debug",propBag.getProperty("debugMode","no")));
			if (propBag.containsKey("lookupDN") && propBag.containsKey("lookupPassword")) {
				lookupCredentials=new Credentials(propBag.getProperty("lookupDN",""),propBag.getProperty("lookupPassword",""));
			}
			if (isDebugMode()) {
				SortedSet ss=new TreeSet(String.CASE_INSENSITIVE_ORDER);
				ss.addAll(propBag.keySet());
				
				StringBuilder sb=new StringBuilder();
				for (Object o : ss) {
					sb.append(o.toString()).append("=[").append(propBag.getProperty(o.toString())).append("]\n");
				}
				String sCheck=propBag.getProperty(CHECK_SSL_CERT_PROP);
				if (sCheck==null) {
					sb.append(CHECK_SSL_CERT_PROP).append(" is not set, using default value of no which disables SSL Certificate Checking for LDAP/SSL!\n");
				} else {
					sb.append("SSL Certificates will ");
					if (!"yes".equalsIgnoreCase(sCheck)) {
						sb.append("NOT ");
					}
					sb.append("be validated for LDAP/SSL\n");
				}
				WebApp.log(STARS+"ldap.properties Contents"+STARS+sb.toString()+STARS);
				WebApp.log("Using LDAP Port="+getLDAPPort());
			}
			if (propBag.getProperty("TrustStorePath")!=null) {
				String sPath=propBag.getProperty("TrustStorePath","sslkey.keystore");
				File f=new File(sPath);
				if (!(f.exists() && f.canRead())) {
					throw new java.io.FileNotFoundException(sPath);
				}

				// the default will be /JRE/lib/security/cacerts
				System.setProperty("javax.net.ssl.trustStore",sPath);
			}

			if (propBag.getProperty("TrustStorePassword")!=null) {
				System.setProperty("javax.net.ssl.trustStorePassword",propBag.getProperty("TrustStorePassword"));
			}

			

			if (guessUserDN()) {
				if (propBag.getProperty("SuffixUserWith")==null || propBag.getProperty("SuffixUserWith").length()==0) {
					WebApp.log(STARS+"\nWARNING:\n\nIf you're guessing the Distinguished Name, you might need SuffixUserWith\n"+STARS);
				}
			} else {
				String s=propBag.getProperty("DefaultContext");
				if (StringUtil.isEmpty(s)) {
					WebApp.log(STARS+"\nWARNING:\n\nYou need to set the default context!\n"+STARS);
				}
				s=propBag.getProperty("lookupDN");
				if (StringUtil.isEmpty(s)) {
					WebApp.log(STARS+"\nERROR:\n\nGuessDN is set to no, which means the system will do a query to\ndetermine the DN, but the lookupDN name is blank so GuessDN will effectively be yes for the initial authenticate bind phase.!\n"+STARS);
				}
				s=propBag.getProperty("searchField");
				if (StringUtil.isEmpty(s)) {
					WebApp.log(STARS+"\nERROR:\n\nIf you're looking up the DN, you need to specify searchField to say what to search!\n"+STARS);
				}

			}

		} catch (IOException e) {
			WebApp.log("AuthProviderLDAP.init()!",e);
		}
	}

	private BoundConnection bindConnection(String ldapHost,Credentials credentials) {

		UnboundConnection uc=getUnboundConnection();

		BoundConnection bc=new BoundConnection(uc,ldapHost,credentials);
		if (isDebugMode()) {
			WebApp.log("bindConnection() "+bc.toString());
		}
		// connect to the server
		try {
			bc.lc.connect( bc.serverName, getLDAPPort());    
		} catch (LDAPException leConnect) {
			bc.bindException=leConnect;
			if (leConnect.getResultCode()==LDAPException.CONNECT_ERROR) {
				WebApp.log("LDAP Connect Error. Check SSL Configuration!");
			}
			if (isDebugMode()) {
				WebApp.log("Error connecting to LDAP Host="+bc.toString(),leConnect);
			} else {
				WebApp.log("Error connecting to LDAP Host="+bc.serverName);
			}
			bc.disconnect();
			return bc;
		}

		// bind to the server
		try {
			bc.lc.bind( ldapVersion, bc.credentials.userName, bc.credentials.password==null ? null : bc.credentials.password.getBytes(LDAP_CHARSET) );
		} catch (LDAPException le) {
			switch (le.getResultCode()) {
			case LDAPException.NO_SUCH_OBJECT:
				bc.additionalErrorInfo="Error: No such entry";
				break;
			case LDAPException.INSUFFICIENT_ACCESS_RIGHTS:
				bc.additionalErrorInfo="Error: Insufficient rights";
				break;
			case LDAPException.ATTRIBUTE_OR_VALUE_EXISTS:
				bc.additionalErrorInfo="Error: Attribute or value exists";
				break;
			case LDAPException.INVALID_CREDENTIALS:
				bc.additionalErrorInfo="Error: Invalid Credentials";
				break;
			case LDAPException.CONNECT_ERROR:
				bc.additionalErrorInfo="Connection Error. Possible SSL Configuration Problem.";
				break;
			default:
				bc.additionalErrorInfo="Error bcCode=["+le.getResultCode()+"] "+le.resultCodeToString();
				break;
			}
			bc.bindException=le;
			bc.disconnect();
			if (isDebugMode()) {
				WebApp.log("Error binding to tree: "+bc.toString());
			} else {
				WebApp.log("Error binding to tree. "+bc.additionalErrorInfo);
			}
		} catch (java.io.UnsupportedEncodingException uee) {
			WebApp.log("Unsupported encoding = ["+LDAP_CHARSET+"]",uee);
		}
		return bc;
	}


	public LoginResult authenticateUser(String UserID, String Password,String TransFormSeed,ResourceBundle rb,Locale lc) {
		lr=new LoginResult();
		return lr;
	}

	private groupusers getGroup(Vector<groupusers> vCurrent,groups grp){
		for (groupusers gu : vCurrent) {
			if (gu.group_id==grp.group_id) {
				return gu;
			}
		}
		return null;
	}

	private void reconcileGroupMemberships(BoundConnection bc,User u,String login){
		List<groups> ldapEquivs=groups.getLDAPGroups();
		if (ldapEquivs.isEmpty()) {
			return;
		}
		Vector<groupusers> vCurrent=u.getUserGroups();
		Set<String> setGroups=getGroups(bc,login,propBag.getProperty("GroupAttributeName","memberOf"));
		Set<groups> setPresent=new HashSet<groups>();
		/* Add the user to any groups that are mapped that they were not in.
		 
			First, iterate over each group we are a member of		 
		 */
		for (String sGroup : setGroups) {
			for (groups g : ldapEquivs) {
				/*
					Now, iterate over each ldap equivalent group, and if the ldap equivalent
					group matches the user group, add the user to that group and mark that
					group as present.
				*/
				groups gFound=null;
				if (g.ldap_equivs.indexOf(';')>0) {
					String[] aEquivs=g.ldap_equivs.split(";");
					for (String sEquivalent : aEquivs) {
						if (sGroup.matches(sEquivalent.trim()) || sGroup.equalsIgnoreCase(sEquivalent.trim())) {
							gFound=g;
							break;
						}
					}
				} else {
					if (sGroup.matches(g.ldap_equivs) || sGroup.equalsIgnoreCase(g.ldap_equivs)) {
						gFound=g;
					}
				}
				if (gFound!=null) {
					if (getGroup(vCurrent,gFound)==null &&
						!setPresent.contains(gFound)
					   ) {
						//	They are not a member of that group.
						groupusers gu=new groupusers();
						gu.group_id=gFound.group_id;
						gu.user_id=u.user_id;
						try {
							gu.update();
						} catch (SQLException se) {
							WebApp.log("Mapping Groups!",se);
						}
					}
					setPresent.add(gFound);
				}
			}
		}
		/*
			Now, if there are groups in the map that they were not LDAP members of,
			remove them.
		*/
		for (groups gTest : ldapEquivs) {
			groupusers gu=getGroup(vCurrent,gTest);
			if (gu!=null && !setPresent.contains(gTest)) {
				try {
					groupusers.delete(gu.group_id,gu.user_id);
				} catch (SQLException se2) {
					WebApp.log("Deleting Groups!",se2);
				}
			}
		}

	}

	/**
	 * Crack the LDAP Server Names, and if a default is passed, make it the 1st element in the array.
	 * 
	 * @author gsexton (7/12/2012)
	 * 
	 * @param serverNames 
	 * @param sDefault 
	 * 
	 * @return String[] 
	 */
	private static String[] setupLDAPServers(String serverNames,String sDefault){
		String[] aResult=serverNames.split(",");
		if (sDefault!=null && aResult.length>1) {
			for (int i=0; i < aResult.length; i++) {
				if (sDefault.equalsIgnoreCase(aResult[i])) {
					if (i>0) {
						/*
							Make the default the zeroth element, and the zeroth some offset.
						*/
						String sTemp=aResult[0];
						aResult[0]=aResult[i];
						aResult[i]=sTemp;
					}
					break;
				}
			}
		}
		return aResult;
	}

	/**
	 * Extract the Connect Daily user account name from the supplied 
	 * user name. For example, given domain\\username with searchField=SAMAccountName, 
	 * return username. 
	 * 
	 * @author gsexton (7/16/2012)
	 * 
	 * @param UserName 
	 * 
	 * @return String 
	 */
	private static String extractUserName(String UserName) {
		String sExpression=propBag.getProperty("ExtractUserNameExpression"),
			sSearchField=propBag.getProperty("searchField",UPN),
			sResult;
		if (sExpression==null) {
			if (UPN.equalsIgnoreCase(sSearchField)) {
				sExpression="(.*)@";
			} else if (SAMACCOUNTNAME.equalsIgnoreCase(sSearchField)) {
				sExpression=".*\\\\(.*)";
			} else {
				sExpression=RegexConstants.CI_FLAG+"cn=([^,]*)";
			}
		}
		if (sExpression==null || "disable".equalsIgnoreCase(sExpression)) {
			sResult=UserName;
		} else {
			Pattern p=Pattern.compile(sExpression);
			Matcher m=p.matcher(UserName);
			if (m.find()) {
				sResult=m.group(1);
			} else {
				sResult=UserName;
			}
		}
		if (isDebugMode()) {
			WebApp.log("extractUserName(\""+UserName+"\") Regular Expression=\""+sExpression+"\" return value: ["+sResult+"]");
		}
		return sResult;
	}

	/** Perform the actual user authentication against the LDAP
	 * directory.
	 * @param UserID The user name to authenticate
	 * @param Password The user supplied password
	 * @param rb ResourceBundle for formatting messages.
	 * @param loc Locale for formatting messages.
	 */
	public LoginResult authenticateUser(String UserID, String Password,ResourceBundle rb,Locale loc){
		final String sPresentedUserID=UserID;
		lr=new LoginResult();
		if (Password==null) {
			lr.getFatalList().add(rb.getString("InvalidUserOrPW"));
			return lr;
		}
		Credentials authenticateCredentials=new Credentials(guessUserDN() ? guessDN(UserID) : UserID, Password);

		/*
			For the connect daily user account name, get the undecorated user name portion.
		*/
		User u=User.getUserByNameCI(extractUserName(UserID).toLowerCase(loc));
		if (isDebugMode()) {
			if (u==null) {
				WebApp.log("Did not find user account.");
			} else {
				WebApp.log("Found User ID: "+u.user_id+" User Name="+u.user_name);
			}
		}
		
		if (UserID.indexOf('\\')>0 && SAMACCOUNTNAME.equalsIgnoreCase(propBag.getProperty("searchField"))) {
			/*
				This is a domain style user name. It's OK for the bind, but not for the query part or user name.
			*/
			UserID=extractUserName(UserID);
		}

		lr.setUser(u);

		String[] aHosts=setupLDAPServers(propBag.getProperty("ServerName","localhost"),
										 u==null ? null : u.getUserPreference(DEFAULT_SERVER_PREFNAME,null)
										);

		
		if (!guessUserDN() && lookupCredentials!=null) {
			String loginDN=null;
			BoundConnection lookupBC=bindConnection(aHosts[0],lookupCredentials);
			if (lookupBC.isBound()) {
				loginDN=lookupBC.getDN(UserID);
				lookupBC.disconnect();
			}
			
			if (loginDN==null || loginDN.length()==0) {
				if (isDebugMode()) {
					WebApp.log("authenticateUser(), lookupDN call failed.\n lookupBC="+lookupBC.toString());
				}
				lr.getFatalList().add(rb.getString("DNNotFound"));
				return lr;
			} else {
				authenticateCredentials.userName=loginDN;
			}
		}

		for (String ldapHost : aHosts) {
			String sTemplateUser=propBag.getProperty(ldapHost+".AutoAddUserTemplate",WebApp.getProperty("Security.AutoAddUserTemplate"));
			if (u==null && sTemplateUser==null) {
				//
				//	We didn't find the user, and there's no auto-add template.
				//
				if (isDebugMode()) {
					WebApp.log("authenticateUser() - The user account ("+UserID+") was not found in the Connect Daily users table, and there is no auto-create template so login using "+ldapHost+" fails.");
				}
				continue;
			}

			lr=authenticateUserOnHost(ldapHost,UserID,sTemplateUser,authenticateCredentials,rb,loc);
			if (lr.getUser()!=null && lr.getFatalList().isEmpty()) {
				User uTemp=lr.getUser();
				uTemp.setUserPreference("AuthenticationUserID",sPresentedUserID);
				uTemp.saveUserPreferences();
				try {
					uTemp.update();
				} catch (SQLException seIgnore) {	//NOPMD
				}
				return lr;
			}
		}
		if (lr.getFatalList().isEmpty()) {
			lr.getFatalList().add(rb.getString("InvalidUserOrPW"));
		}
		return lr;
	}

	private LoginResult authenticateUserOnHost(
											  String ldapHost, 
											  String UserID,
											  String sTemplateUser,
											  Credentials authenticateCredentials,
											  ResourceBundle rb,
											  Locale loc){
		User u=lr.getUser();
		lr.reset();

		BoundConnection bc=bindConnection(ldapHost,authenticateCredentials);
		if (isDebugMode()) {
			WebApp.log("AuthProviderLDAP.authenticateUserOnHost bindConnection result="+bc.toString());
		}

		if (!bc.isBound()) {
			lr.getFatalList().add(rb.getString("InvalidUserOrPW"));
			return lr;
		}
		
		/*
			OK, we're bound using the credentials, let's make sure we can find our own
			distinguished name. This can happen in a multi-domain/server environment.
		*/
		String loginDN=bc.getDN(UserID);

		if (loginDN==null || loginDN.length()==0) {
			//
			// Didn't find.
			//
			lr.getFatalList().add(rb.getString("DNNotFound"));
			bc.disconnect();
			return lr;
		}

		if (u==null) {
			if (isDebugMode()) {
				WebApp.log("We found the user in the tree, but we did not find the user in\r\n"+
						   "The Connect Daily user table. If auto-create is not enabled, this\r\n"+
						   "login will fail. If you don't have auto-create enabled, you have\r\n"+
						   "to create the user in Connect Daily first before they can login."
						  );

			}
			// OK, we've found the distinguished name in the system, but
			// we have not found the user account. Let's see if we can auto-create it.
			// 
			// The precondition is checked in the calling routine.
			//
			if (isDebugMode()) {
				WebApp.log("Auto Creating User="+UserID+" using Template="+sTemplateUser);
			}
			u=User.createUserFromTemplate(extractUserName(UserID),sTemplateUser);
			lr.setUser(u);
			try {
				String sTemp=getFullName(bc,UserID);
				if (!StringUtil.isEmpty(sTemp)) {
					u.full_name=sTemp.trim();
				}
				sTemp=getEmailAddress(bc,UserID);
				if (!StringUtil.isEmpty(sTemp)) {
					u.email_address=sTemp.toLowerCase(loc).trim();
				}
				u.setUserPreference(DEFAULT_SERVER_PREFNAME,ldapHost);
				u.saveUserPreferences();
				u.update();
				reconcileGroupMemberships(bc,u,UserID);
			} catch (Exception e) {
				WebApp.log("Creating user from template!",e);
			}
		} else {
			if (verifyLoginPreconditions(u,rb).isEmpty()) {
				if (isDebugMode()) {
					WebApp.log("verifyLoginPreconditions() return isEmpty()!");
				}
				String sTemp=getFullName(bc,UserID);
				if (!StringUtil.isEmpty(sTemp)) {
					u.full_name=sTemp.trim();
				}
				sTemp=getEmailAddress(bc,UserID);
				if (!StringUtil.isEmpty(sTemp)) {
					u.email_address=sTemp.toLowerCase(loc).trim();
				}
				u.setUserPreference(DEFAULT_SERVER_PREFNAME,ldapHost);
				u.saveUserPreferences();
				reconcileGroupMemberships(bc,u,UserID);
				processSuccessfulLogin(u,rb,loc);	// this implicitly does an update()
			}
		}
		return lr;
	}


	public Set<String> getGroups(BoundConnection bc,String UserID,String setAttributeName){
		Set<String> setResult=bc.getDSAttributeSet(UserID,setAttributeName);
		if (isDebugMode()) {
			StringBuilder sb=new StringBuilder(256);
			sb.append("Group Memberships for: "+UserID+" Using Attribute Name:"+setAttributeName).append("\n\n");
			TreeSet<String> ts=new TreeSet(String.CASE_INSENSITIVE_ORDER);
			ts.addAll(setResult);
			for (String sGroup : ts) {
				sb.append(' ').append(sGroup).append('\n');
			}
			WebApp.log(sb.toString());
		}
		return setResult;
	}

	public String getFullName(BoundConnection bc,String UserID){
		return bc.getDSAttribute(UserID,propBag.getProperty("fullName"));
	}

	public String getEmailAddress(BoundConnection bc,String UserID){
		return bc.getDSAttribute(UserID,propBag.getProperty("emailAddress"));
	}
	/**
	 * Initialize a user from the directory. This depends on the user being distinguished enough, and 
	 * that there is a lookup account available. 
	 * 
	 * @param u 
	 * 
	 * @return User 
	 */
	public User initUserFromDirectory(User u) {

		String[] aHosts=setupLDAPServers(propBag.getProperty("ServerName","localhost"),
										 u==null ? null : u.getUserPreference(DEFAULT_SERVER_PREFNAME,null)
										);
		BoundConnection bc=bindConnection(aHosts[0],lookupCredentials);
		if (bc.isBound()) {
			String s=getFullName(bc,u.user_name);
			if (s!=null) {
				u.full_name=s;
			}
			s=getEmailAddress(bc,u.user_name);
			if (s!=null) {
				u.email_address=s;
			}
			bc.disconnect();
		}
		return u;
	}


	@SuppressWarnings("PMD.SystemPrintln")
	public static void main(String[] argv) throws IOException {
		if (argv.length!=4) {
			System.out.println("Usage: AuthProviderLDAP configuration.properties ldap.properties userID Password");
			return;
		}
		setDebugMode(true);
		File fProperties=new File(argv[0]);
		WebApp.setPropertiesFile(fProperties);

		WebApp.setProperty("AuthenticationProvider","com.mhsoftware.cdaily.support.security.AuthProviderLDAP");
		AuthProviderLDAP auth=(AuthProviderLDAP)WebApp.getAuthenticationProvider();
		init(argv[1]);
		LoginResult lr=auth.authenticateUser(argv[2],argv[3],com.mhsoftware.cdaily.servlet.CalServlet.getResourceBundle(Locale.ENGLISH),java.util.Locale.getDefault());
		System.out.println("Trying user ID=["+argv[2]+"]");
		System.out.println("authenticateUser() returned: "+lr+"\nAdditional info is: "+auth.getFailureInformation());
		System.out.println("Full Name for "+argv[2]+" is ["+auth.getFullName(argv[2])+"]");
		System.out.println("Email Address for "+argv[2]+"=["+auth.getEmailAddress(argv[2])+"]");
	}
}




