package com.mhsoftware.cdaily.support.security;

/*
	Connect Daily Web Calendar Software

	Copyright 2002-2005, MH Software, Inc. All Rights Reserved
	5023 W 120th Ave #311, Broomfield CO 80020
	+1 303 438 9585

	Last Edit By: "$Author: gsexton $"
	Last Checkin: "$Date: 2014/09/16 22:16:36 $"
	Revision #  : "$Revision: 1.18 $"

*/

import java.sql.SQLException;
import java.sql.Timestamp;
import java.text.DateFormat;
import java.text.MessageFormat;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.TimeZone;
import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import com.MHSoftware.times.TimeConstants;

import com.mhsoftware.cdaily.business.User;
import com.mhsoftware.cdaily.support.WebApp;

/**
*	Provides base class for an authentication provider
*	plug-in architecture for Connect Daily.
*
*	<BR><BR>
*	For example, you can write an LDAP authentication
*	module that extends and overrides this class.
*	<BR><BR>
*
*	To use your class as the authentication provider,
*	Add an authentication provider line to the
*	configuration.properties file:<BR><BR>
*	E.G.<BR>
*	<CODE>AuthenticationProvider=com.mycompany.myclass</CODE>
*	<BR><BR>
*	@author George Sexton
*/
public abstract class AuthenticationProvider {

	public static final int PLAINTEXT_PW_AVAILABLE=1;
	public static final int DIRECTORY_AVAILABLE=2;
	public static final int CHANGES_PASSWORDS=4;
	public static final int DO_PRETRANSFORM=8;
	/** This flag should be set if the provider depends upon the
	 *  servlet container to do authentication.
     */
	public static final int CONTAINER_SECURITY=16;
	/**
	 * This flag should be set if the provider can automatically
	 * create users once they are authenticated to an external
	 * source like LDAP/AD or the container.
     */
	public static final int DOES_AUTO_ADD=32;
	/** This provider is a single-signon provider. There's something magic in
	 *  the request that the auth provider is going to pick up and use.
	 */
	public static final int SINGLE_SIGNON=64;

	protected static ServletContext context=null;
	protected static int provider_flags;
	protected LoginResult lr;
	private static boolean debugMode=false;
	
	/**
	 * This initialization method is required to allow the
	 * AuthenticationProvider to set it's environment up.
	 * <BR>
	 * @param sc	Providing the servlet context allows us
	 * the class to use getResourceAsStream(), and also provide
	 * logging back to the log file for the context.
	 *
	 */

	public static void init(ServletContext sc){
		context=sc;
		provider_flags=0;
	}

	public static boolean isDebugMode(){
		return debugMode;
	}

	protected static void setDebugMode(final boolean mode){
		debugMode=mode;
	}
	protected static void setDebugMode(final String mode) {
		if (mode==null) {
			setDebugMode(false);
		} else {
			char c=mode.charAt(0);
			setDebugMode(c=='y' || c=='Y' || c=='1' || c=='t' || c=='T' || "on".equalsIgnoreCase(mode));
		}
	}
	/**
	*	Method to authenticate a user.
	*
	* @param	UserID		The supplied User ID to authenticate.
	* @param	Password 	The supplied password to use.
	* @param	rb 			Resource Bundle for formatting messages.
	* @param	lc			Locale used for formatting times and dates.
	*
	*	@return	If the user is successfully authenticated, this method
	*	returns the name of the Connect Daily user. If authentication
	*	fails, this method should return null.
	*/
	public abstract LoginResult authenticateUser(String UserID, String Password,ResourceBundle rb,Locale lc);

	public abstract LoginResult authenticateUser(String UserID, String Password,String TransFormSeed,ResourceBundle rb,Locale lc);

	public LoginResult authenticateUser(HttpServletRequest req,ResourceBundle rb,Locale lc) {
		throw new IllegalStateException("Method not implemented!");
	}

	/**
	*	Optionally, implement this method in such a way to provide
	*	additional authentication failure information to the user.
	*	<BR><BR>
	*	If you don't want to provide additional information, return null.
	*/
	public  String getFailureInformation(){
		StringBuffer sbResult=new StringBuffer();
		if (lr!=null) {
			Iterator itr=lr.getFatalList().iterator();
			while (itr.hasNext()) {
				sbResult.append((String)itr.next()).append("<BR>\n");
			}
			itr=lr.getWarningList().iterator();
			while (itr.hasNext()) {
				sbResult.append((String)itr.next()).append("<BR>\n");
			}
		}
		return sbResult.toString();
	}

	/** This method is called by the password change screen, to apply
	 * any authenticator specific transforms to the user password.
	 * An example where this would be used is an authenticator
	 * that stores the password as an MD5 hash, rather than clear
	 * text. If you don't want to do a transformation, just return
	 * Password unmodified.
	 * @param Password	The password to transform.
	 * @return The transformed password. This may be the password
	 * totally unmodified, or a hash value of the password.
	 *
	 */
	public  String transformPassword( final String Password) {
		if (Password==null) {
			return Password;
		} else {
			return Password.trim();
		}
	}


	/** If this provider is a DS Provider, read the
	 * full name of the user from the directory.
	 *
	 * @param UserID the ID of the user to get the full name for.
	 */
	public String getFullName(String UserID){
		if (lr==null || lr.getUser()==null) {
			return null;
		} else {
			return lr.getUser().full_name;
		}
	}

	/** If this provider is a DS Provider, read the
	 * Email Address of the user from the directory.
	 * @param UserID the ID of the user to get the full name for.
	 */
	public  String getEmailAddress(String UserID){
		if (lr==null || lr.getUser()==null) {
			return null;
		} else {
			return lr.getUser().email_address;
		}
	}

	/**
	 * Return TRUE if a provider flag is set, false
	 * otherwise.
	 *
	 * @param flag The flag to check.
	 */
	public boolean isProviderFlagSet(final int flag){
		return((provider_flags & flag)==flag);
	}

	/**
	 * This method returns the various provider flags that
	 * the caller can use to inspect the provider.
	*/
	public int getProviderFlags(){
		return provider_flags;
	}



	/**
	 * Process housekeeping steps associated with a successful login.
	 * The return list will be a series of warnings or messages
	 * <BR><BR>
	 * This will handle things like warning users that their password
	 * has expired, and how many grace logins remain.
	 * <BR><BR>
	 * Any returned messages are advisory in nature and should be
	 * displayed to the user. HOWEVER, THIS ROUTINE MAY TOGGLE THE
	 * ENABLED FLAG IF GRACE LOGINS ARE ALL USED UP, SO THE CALLING
	 * ROUTINE SHOULD CHECK THE ENABLED FLAG AFTER THIS CALL.
	 * <BR><BR>
	 * @param u The user who logged in.
	 * @param rb Resource Bundle to use for messages.
	 * @param lc Locale used for formatting date and time messages.
	 */
	public List processSuccessfulLogin(User u,ResourceBundle rb,Locale lc){
		if (lr==null) {
			lr=new LoginResult();
		}
		List alResult = lr.getWarningList(); 
		// Reset the bad-login count to 0
		u.blogincount=0;
		u.saveLastSignin();
		if (u.pwchangedate==null) {
			u.pwchangedate=new java.sql.Timestamp(System.currentTimeMillis());
		}
		if (u.getUserPreference("pwnonce",null)!=null) {
			u.setUserPreference("pwnonce",null);
			u.saveUserPreferences();
		}
		final long iPWDays = WebApp.getIntProperty("Security.MaximumPasswordAge", -1); 
		if (iPWDays>0) {
			MessageFormat mf;
			// We have password aging turned on
			Timestamp tChange=new Timestamp(
										   u.pwchangedate.getTime()+(TimeConstants.MILLISECONDS_DAY*iPWDays)
										   );
			if (System.currentTimeMillis()>tChange.getTime()) {
				u.grlogrem--;

				if (u.grlogrem<0) {
					u.enabled=User.ACCT_DISABLED;
					alResult.add(rb.getString("Security.AccountLocked"));
				}
				//
				// Add a notice telling the user the # grace logins remaining.
				//
				mf=new MessageFormat(rb.getString("Security.PasswordExpired"));
				alResult.add(mf.format(new Object[] {
					u.grlogrem>0 ? Integer.valueOf(u.grlogrem) : Integer.valueOf(0)
					},new StringBuffer(),null).toString());

			} else {
				//
				// Since the PW is not expired, let's see if it's close to expiring
				// and warning the user if so.
				//
				final long iWarnDays=WebApp.getIntProperty("Security.WarnPasswordDays",7);
				final long warnTime=tChange.getTime()-(iWarnDays*TimeConstants.MILLISECONDS_DAY);
				if (System.currentTimeMillis()>warnTime) {
					DateFormat df=WebApp.getDateTimeFormat(DateFormat.LONG,DateFormat.SHORT,lc);
					df.setTimeZone(u.tz_string==null ? WebApp.getTimeZone() : TimeZone.getTimeZone(u.tz_string));
					// Write a message telling them when their password will expire.
					mf=new MessageFormat(rb.getString("Security.PasswordExpiresSoon"));
					alResult.add(mf.format(new Object[] {df.format(tChange)},new StringBuffer(),null).toString());
				}
			}
		}
		try {
			u.update();
		} catch (SQLException e) {
			WebApp.log("Updating user",e);
		}
		return alResult;
	}

	/**
	 * Verify pre-conditions for login are met. These preconditions are:
	 * <OL>
	 * <LI>Account not disabled.
	 * <LI>Bad Login Count not Exceeded.
	 * </OL>
	 * <BR><BR>
	 * A return of any messages is a signal that pre-conditions are not met
	 * and login should not be attempted.
	 * <BR><BR>
	 * As a side effect, if the period for bad login count retention has
	 * passed, then this will clear the bad login count.
	 *
	 * @param u The user to verify pre-conditions for
	 * @param rb The resource bundle for message formatting.
	 *
	 */
	public List verifyLoginPreconditions(User u, ResourceBundle rb){
		List alResult=lr.getFatalList();
		if (!u.isEnabled()) {
			alResult.add(rb.getString("Security.AccountLocked"));
			return alResult;
		}

		final int iLockCount=WebApp.getIntProperty("Security.BadLoginLockCount",Integer.MIN_VALUE);
		if (iLockCount>0) {
			// Bad Login Locking is enabled. Check that we haven't exceeded our count.
			if (u.lastblogin!=null) {
				if (u.blogincount>=iLockCount) {
					final int iLockoutPeriod=WebApp.getIntProperty("Security.AccountLockoutPeriod",Integer.MIN_VALUE);
					if (System.currentTimeMillis()>
						u.lastblogin.getTime()+(TimeConstants.MILLISECONDS_MINUTE*iLockoutPeriod)) {
						u.blogincount=0;
						try {
							u.update();
						} catch (SQLException e) {
							WebApp.log("Updating user",e);
						}
					}
				} else {
					// The acount is not locked.
					// if the login retention period has lapsed, clear the count.
					final int iPeriod=WebApp.getIntProperty("Security.BadLoginPeriod",Integer.MIN_VALUE);
					if ((iPeriod > 0) &&
						(u.blogincount>0) &&
						(System.currentTimeMillis()>
						 u.lastblogin.getTime()+(TimeConstants.MILLISECONDS_MINUTE*iPeriod))
					   ) {
						// The bad login retention period has passed, clear the bad login count.
						u.blogincount=0;
						try {
							u.update();
						} catch (SQLException e) {
							WebApp.log("Updating user",e);
						}
					}
				}
			}
			if (u.blogincount>=iLockCount) {
				alResult.add(rb.getString("Security.AccountLocked"));
			}
		}
		return alResult;
	}

	/**
	 * Process housekeeping steps associated with a failed login.
	 * The return list will be a series of warnings or messages.
	 *
	 * <BR><BR>
	 * A return of any messages should be displayed to the user.
	 *
	 * @param u The user account the login failed for.
	 * @param rb The resource bundle to use for message formatting.
	 */
	public List processFailedLogin(User u, ResourceBundle rb){
		List alResult=lr.getFatalList();

		u.lastblogin=new java.sql.Timestamp(System.currentTimeMillis());
		u.blogincount++;
		final int iBadLoginLockCount=WebApp.getIntProperty("Security.BadLoginLockCount",Integer.MAX_VALUE);
		if ((iBadLoginLockCount>0) && (u.blogincount>=iBadLoginLockCount)) {
			if (WebApp.getIntProperty("Security.AccountLockoutPeriod",-1)==-1) {
				u.enabled=User.ACCT_DISABLED;
			}
			alResult.add(rb.getString("Security.AccountLocked"));
		}
		if (alResult.isEmpty()) {
			// It's a failure but we don't have a reason yet. It must be invalid pw.
			alResult.add(rb.getString("InvalidUserOrPW"));
		}
		try {
			u.update();
		} catch (SQLException e) {
			WebApp.log("Updating user",e);
		}
		return alResult;
	}
}
