import java.io.*;
import java.util.*;
import com.mhsoftware.cdaily.support.CSSParser;
import com.MHSoftware.dates.SaneDate;

public class CSSFixerUpper {

    private static String[][] aColorData = {
        { "clrPageColor", null, "body", "background-color" }, { "clrTextColor", null, "body", "color" }, { "clrHeadingBGColor", null, "th", "background-color" }, { "clrHeadingTextColor", null, "th", "color" }, { "clrCalendarBGColor", null, "td.mhvymonthcell", "background-color" }, { "clrEventLinkColor", null, "A", "color" }, { "clrCalendarTextColor", null, "body", "color" }, { "clrDayAndTimeLabels", null, ".MHVCDayNumeral", "color" }, { "clrBorderColor", null, "table.mhvctable", "border-color" }, { "clrCurrentDayBorder", null, "td.mhvctodaycell", "border-color" }, { "clrShadedCell", "table.mhvctable", "td.offmonthcell", "background-color" }
    };
    private static int[] fallbackColors = {
        4, 5, 0, 1, 0, 1, 2, 1, 2, 8, 0
    };

    /**
     * A class to fix up customer CSS files and add new Connect Daily CSS 
     * <BR><BR> 
     * What this class will do is read a customer's CSS file, and then 
     * read a snippet of new CSS. Color substitutions are made on the 
     * snippet, using values from the customer CSS file, and the 
     * snippet is then inserted into the appropriate position in the customer's 
     * CSS file. 
     * 
     * @author gsexton (4/1/2011)
     * 
     * @param argv 
     *  
     * Arguments Are: 
     *  
     * --header=file_name.css <br>
     * --footer=file_name.css <br>
     * ... Files To Process. Each Modified file will be saved as 
     * file.css.save-yyyymmdd. 
     *  
     * --header is assumed to be imports and won't be parsed for color replacement. 
     */
    public static void main(String[] argv) throws IOException {
        if (argv.length < 2) {
            System.err.println("Usage: CSSFixerUpper [--header=file_name.css --footer=file_name.css] file_to_be_modified.css");
            System.err.println("Refer to the class JavaDoc for more info.");
            System.exit(0);
        }
        String sFooterSnippet = null, sHeaderSnippet = null;
        // readSnippet(argv[0]);
        SaneDate sd = new SaneDate();
        CSSParser parseSnippet = null;
        String saveSuffix = ".save-" + sd.toString(SaneDate.FORMAT_YYYY_MM_DD);
        for (String sArg: argv) {
            if (sArg.startsWith("--header=")) {
                String sHeader = sArg.substring(sArg.indexOf('=') + 1);
                System.err.println("Reading Header File=" + sHeader);
                sHeaderSnippet = readSnippet(sHeader);
            } else if (sArg.startsWith("--footer=")) {
                String sFooter = sArg.substring(sArg.indexOf('=') + 1);
                System.err.println("Reading Footer File=" + sFooter);
                sFooterSnippet = readSnippet(sFooter);
                parseSnippet = CSSParser.parseFile(sFooter);
            } else {
                if (sFooterSnippet == null && sHeaderSnippet == null) {
                    throw new IllegalStateException("No header or footer snippet specified for processing!");
                }
                File f = new File(sArg);

                if (f.exists()) {
                    saveFile(sArg, saveSuffix);
                    processCSSFile(sArg, sHeaderSnippet, sFooterSnippet, parseSnippet);
                } else {
                    System.err.println("File: " + sArg + " does not exist! Ignoring!");
                }
            }
        }
    }


    private static String readSnippet(String snippetFile) throws IOException {
        StringBuilder sb = new StringBuilder(2048);
        char[] buf = new char[512];
        FileInputStream fis = new FileInputStream(snippetFile);
        InputStreamReader rdr = new InputStreamReader(fis, "UTF-8");
        int iRead;
        while ((iRead = rdr.read(buf, 0, buf.length)) > 0) {
            sb.append(buf, 0, iRead);
        }
        rdr.close();
        fis.close();

        return sb.toString();
    }

    private static void copyFile(File fInput, File fOutput) throws IOException {
        FileInputStream fis = new FileInputStream(fInput);
        FileOutputStream fos = new FileOutputStream(fOutput);
        byte[] buf = new byte[2048];
        int iRead;
        while ((iRead = fis.read(buf)) > 0) {
            fos.write(buf, 0, iRead);
        }
        fis.close();
        fos.close();
    }

    private static void saveFile(String original, String saveSuffix) throws IOException {
        
        String sSave = original + saveSuffix;
        File f = new File(sSave);
        if (f.exists()) {
            f.delete();
        }
        System.err.println("Saving file "+original+" to "+f.getName());
        copyFile(new File(original), f);
    }

    private static void processCSSFile(String fileName,
                                       String sHeaderSnippet,
                                       String sFooterSnippet,
                                       CSSParser parserSnippet) throws IOException {

        CSSParser p = CSSParser.parseFile(fileName);

/* 
        Set<String> setKeys=parserSnippet.getElements().keySet();
        Map<String,Map<String,String>> mElements=p.getElements();
 
        for (String sCheck : setKeys) {
            if (mElements.containsKey(sCheck)) {
                System.out.println("Element: "+sCheck+" found in destination file "+fileName+". Skipping.");
                return;
            }
        }
*/
        String sReplace = "\n" + sFooterSnippet;
        boolean bReplace = false;
        for (int i = 0; i < aColorData.length; i++) {
            String[] aThisColor = aColorData[i];
            // Color Data
            // 0 - color name
            // 1 - Container name
            // 2 - element name
            // 3 - attribute name

            String sColor = p.getNestedElementAttributeValue(aThisColor[1], aThisColor[2], aThisColor[3]);
            if (sColor == null) {
                aThisColor = aColorData[fallbackColors[i]];
                sColor = p.getNestedElementAttributeValue(aThisColor[1], aThisColor[2], aThisColor[3]);
            }
            if (sColor != null && sReplace.indexOf(aThisColor[0]) > 0) {
                sReplace = sReplace.replaceAll(aThisColor[0], sColor);
                bReplace |= true;
            }
        }
        String sOut = readSnippet(fileName);
        boolean bWrite = false;
        if (sHeaderSnippet != null) {
            sOut = sHeaderSnippet + "\n" + sOut;
            bWrite = true;
        }
        if (bReplace) {
            bWrite = true;
            sOut += sReplace;
        }
        if (bWrite) {
            OutputStreamWriter osw = new OutputStreamWriter(new FileOutputStream(fileName), "UTF-8");
            osw.write(sOut.replaceAll("\r\n","\n"));
            osw.close();
        }
    }

}
